import time

from loguru import logger
from watchdog.observers import Observer
from watchdog.tricks import ShellCommandTrick

from .threaded import Threaded
from .events import Event


class Watcher(Threaded):
    def __init__(self, config, dispatcher) -> None:

        super().__init__()
        self.config = config
        self.dispatcher = dispatcher
        self.handler = FileChangeHandler(config, dispatcher)
        self.observer = Observer()
        for pathname in set(config.dirs):
            self.observer.schedule(self.handler, pathname, recursive=True)

    def start(self):

        logger.debug("Starting observer thread")
        self.observer.start()
        logger.debug("Observer thred started")

    def stop(self):

        logger.debug("Stopping observer thread")
        self.observer.stop()
        self.observer.join()
        logger.debug("Observer thread stopped")


class FileChangeHandler(ShellCommandTrick):
    def __init__(self, config, dispatcher):

        self.config = config
        self.dispatcher = dispatcher
        self.cooldown_timer = time.time()
        super().__init__(
            shell_command=config.command,
            patterns=config.patterns,
            ignore_patterns=config.ignore_patterns,
            ignore_directories=config.ignore_dirs,
            wait_for_process=True,
            drop_during_process=True,
        )

    def on_any_event(self, event):

        command_result = None

        timeout = time.time() - self.cooldown_timer
        if timeout < self.config.wait:
            logger.debug(
                f"Ignoring watcher event because less time passed then specified in cooldown: {timeout:.2f} < {self.config.wait:.2f}")
            return None

        # Report event processing
        logger.debug(f"Processing event {event}.")

        # Run shell command
        try:
            command_result = super().on_any_event(event)
        except Exception as e:
            logger.error(f"{e}")
        logger.debug(f"Shell command executed with result: {command_result}.")

        # Register file changed event
        self.dispatcher.report(Event.ShellCommandFinished)

        # Remember current time for cooldown timer
        self.cooldown_timer = time.time()

        return command_result



import time
from typing import List

from loguru import logger


class Threaded:
    
    def start(self):
        """Start the thread."""

    def stop(self):
        """Stop the thread."""

    @staticmethod
    def run(runnables: List["Threaded"]):

        try:
            # Start all threads
            for r in runnables:
                r.start()

            # Sleep indefinitely
            while True:
                time.sleep(1)

        except KeyboardInterrupt:
            logger.debug("Stopping all threads on KeyboardInterrupt")

        except Exception as ex:
            logger.error(f"{ex}")

        finally:
            # Stopping all threads
            for r in runnables:
                r.stop()



from enum import Enum

from loguru import logger


class Event(Enum):
    ShellCommandFinished = 1


class EventDispatcher(list):
    def report(self, event):
        """Report new event."""

        self.append(event)
        logger.debug(f"New {event} event has been reported.")
        logger.debug(f"Events: {self}")

    def claim(self, event):
        """Claim all events of certain type."""

        claimed = [e for e in self if e == event]
        unclaimed = [e for e in self if e != event]
        super().clear()
        super().extend(unclaimed)

        # Report to logs
        if claimed:
            logger.debug(
                f"{len(claimed)} {event} events have been claimed from the dispatcher."
            )
            logger.debug(f"Events: {self}")

        return claimed



import asyncio
import pathlib
import importlib.resources as pkg_resources
import threading
from typing import Optional
import webbrowser

import aiofiles
from aiohttp import web
from bs4 import BeautifulSoup
from loguru import logger

from .threaded import Threaded
from .events import Event, EventDispatcher
from . import js


class Server(Threaded):
    def __init__(self, config, dispatcher: EventDispatcher) -> None:

        self.config = config
        self.dispatcher = dispatcher
        self.injection_script = pkg_resources.read_text(js, "wutch.js")

        # Initialize async thread and web server application
        self.event_loop = asyncio.new_event_loop()
        self.app = web.Application()
        self.app.add_routes([
            web.get("/ws", self.handle_websocket),
            web.get(r"/{file:.*}", self.handle_http),
        ])
        self.runner = web.AppRunner(self.app)
        self.thread = threading.Thread(
            target=self.event_loop.run_forever,
            daemon=True
        )

    def start(self):

        self.thread.start()
        # Submit server running task to the event loop
        asyncio.run_coroutine_threadsafe(
            self._start_async_runner(), self.event_loop)
        logger.debug("Server thread started")

        if self.config.index and not self.config.no_browser:
            self._open_browser()

    def stop(self):

        # Event loop should be properly stopped using command below.
        # But I couldn't supress the
        #   "Task was destroyed but it is pending!"
        # error thrown by the websocket handler when the loop got stopped.
        # So, instead, I resorted to the force thread.join with 0 timeout.
        # self.event_loop.call_soon_threadsafe(self.event_loop.stop)
        self.thread.join(timeout=0.0)
        logger.debug("Server thread stopped")

    async def handle_http(self, request: web.Request):

        # Identify path of the requested file
        file = request.match_info["file"]
        logger.debug(f"Handling request: {file}")
        path = pathlib.Path(self.config.build) / file

        if not path.exists():
            logger.error(f"File {path} not found")
            raise web.HTTPNotFound()

        elif path.is_dir():
            file_path = path / "index.html"
            return await self._load(file_path, content_type="text/html", content_modifier=self._inject_script)

        if "htm" in path.suffix.lower():
            return await self._load(path, content_type="text/html", content_modifier=self._inject_script)

        elif "css" in path.suffix.lower():
            return await self._load(path, content_type="text/css")

        elif "js" in path.suffix.lower():
            return await self._load(path, content_type="text/javascript")

        else:
            return await self._load(path, mode="rb", content_type="application/octet-stream")

    async def handle_websocket(self, request):

        logger.debug(f"Handling websocket: {request}")
        ws = web.WebSocketResponse()
        await ws.prepare(request)

        while True:
            if self.dispatcher.count(Event.ShellCommandFinished):
                # Claim all events
                self.dispatcher.claim(Event.ShellCommandFinished)
                await ws.send_str("reload")
                logger.debug("Reload message sent via websocket")

            await asyncio.sleep(1)

        return ws

    async def _start_async_runner(self):

        await self.runner.setup()
        site = web.TCPSite(self.runner, self.config.host, self.config.port)
        await site.start()

    async def _load(self, path, mode="r", content_type="text/plain", content_modifier=None):

        # Read file
        contents = None
        async with aiofiles.open(path, mode=mode) as f:
            contents = await f.read()

        # Call modifier if it was passed
        if content_modifier:
            contents = content_modifier(contents)

        # Return web response
        return web.Response(body=contents, content_type=content_type)

    def _inject_script(self, contents):

        soup = BeautifulSoup(contents, "html.parser")
        if soup.head:
            script_tag = soup.new_tag("script")
            script_tag.string = self.injection_script
            soup.head.append(script_tag)
        contents = soup.prettify()
        return contents.encode("utf-8")

    def _open_browser(self):

        url = f"http://{self.config.host}:{self.config.port}/{self.config.index}"
        logger.debug(f"Opening browser at: {url}")
        webbrowser.open_new_tab(url)



import ilexconf
import argparse

from loguru import logger


class Config(ilexconf.Config):
    defaults = ilexconf.Config(
        # Directories to watch for the file changes.
        dirs=["."],
        # Directories to ignore during watching for file changes.
        ignore_dirs=["_build", "build"],
        # File patterns to watch for changes.
        patterns=["*"],
        # File patterns to ignore.
        ignore_patterns=[],
        # Shell command to run each time files in the watched directories
        # get changed.
        command="sphinx-build",
        # In which directory HTML files built by shell command will appear.
        build="_build/html",
        # File patterns which should be injected with page refreshing
        # javascript.
        inject_patterns=["*.htm*"],
        # Path to file to be opened by webbrowser relative to build.
        index="index.html",
        # Which host and port to bind the internal HTTP server to.
        host="localhost",
        port=5010,
        # Cooldown period after which the command does not run again and ignores events
        # that trigger it.
        wait=1,
        # Do not open browser after launch of wutch.
        no_browser=False,
        # Do not start a webserver.
        no_server=False,
    )

    def __init__(self):

        super().__init__(
            self.defaults,
            ilexconf.from_json("wutch.cfg", ignore_errors=True),
            ilexconf.from_env(prefix="WUTCH_"),
            ilexconf.from_argparse(self._parse_arguments()),
        )

        logger.debug(f"{self}")

    @staticmethod
    def _parse_arguments():

        parser = argparse.ArgumentParser(
            prog="wutch",
            description=(
                "Watches specified directories for the changes in files "
                "mathing given patterns and runs a shell command each time. "
                "Opens/refreshes the webpage after the command is done."
            ),
        )
        parser.add_argument(
            "-c",
            "--command",
            help=f"Shell command executed in response to file changes. Defaults to: {Config.defaults.command}.",
            type=str,
        )
        parser.add_argument(
            "-p",
            "--patterns",
            help=f"Matches paths with these patterns (separated by ' '). Defaults to: {Config.defaults.patterns}.",
            nargs="*",
            type=str,
        )
        parser.add_argument(
            "-P",
            "--ignore-patterns",
            help=f"Ignores file changes in these patterns (separated by ' '). Defaults to: {Config.defaults.ignore_patterns}.",
            nargs="*",
            type=str,
        )
        parser.add_argument(
            "-d",
            "--dirs",
            help=f"Directories to watch (separated by ' '). Defaults to: {Config.defaults.dirs}.",
            nargs="*",
            type=str,
        )
        parser.add_argument(
            "-D",
            "--ignore-dirs",
            help=f"Ignore file changes in these directories (separated by ' '). Defaults to: {Config.defaults.ignore_dirs}.",
            nargs="*",
            type=str,
        )
        parser.add_argument(
            "-w",
            "--wait",
            help=f"Wait N seconds after the command is finished before refreshing the web page. Defaults to: {Config.defaults.wait}.",
            type=int,
        )
        parser.add_argument(
            "-b",
            "--build",
            help=f"Build directory containing files to render in the browser. Defaults to: {Config.defaults.build}.",
            type=str,
        )
        parser.add_argument(
            "-I",
            "--inject-patterns",
            help=f"Patterns of files to inject with JS code that refreshes them on rebuild (separated by ' '). Defaults to: {Config.defaults.inject_patterns}.",
            nargs="*",
            type=str,
        )
        parser.add_argument(
            "-i",
            "--index",
            help=f"File that will be opened in the browser with the start of the watcher. Defaults to: {Config.defaults.index}.",
            type=str,
        )
        parser.add_argument(
            "--host",
            help=f"Host to bind internal HTTP server to. Defaults to: {Config.defaults.host}.",
            type=str,
        )
        parser.add_argument(
            "--port",
            help=f"TCP port to bind internal HTTP server to. Defaults to: {Config.defaults.port}.",
            type=int,
        )
        parser.add_argument(
            "-B",
            "--no-browser",
            help=f"Do not open browser at wutch launch. Defaults to: {Config.defaults.no_browser}.",
            action="store_true"
        )
        parser.add_argument(
            "-S",
            "--no-server",
            help=f"Do not start the webserver, just launch the shell command. Defaults to: {Config.defaults.no_server}.",
            action="store_true"
        )

        return parser.parse_args()



from .app import WutchApplication


__version__ = "1.0.6"


def cli():
    """Script invokation function used by poetry."""

    wutch_app = WutchApplication()
    wutch_app.cli()


if __name__ == "__main__":
    # This gets launched if the wutch package is invoked as a python script:
    #   python -m wutch
    #   python wutch/__init__.py
    cli()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['wutch', 'wutch.js']

package_data = \
{'': ['*']}

install_requires = \
['aiofiles>=0.6.0,<0.7.0',
 'aiohttp>=3.7.3,<4.0.0',
 'beautifulsoup4>=4.9.3,<5.0.0',
 'ilexconf>=0.9.5,<0.10.0',
 'loguru>=0.5.3,<0.6.0',
 'watchdog>=1.0.2,<2.0.0']

entry_points = \
{'console_scripts': ['wutch = wutch:cli']}

setup_kwargs = {
    'name': 'wutch',
    'version': '1.0.6',
    'description': 'Watch and rebuild your files like LiveServer',
    'long_description': '# Wutch\n\n`wutch` watches for changes in the directories and runs a shell command for\nevery change. It can also open a browser a display whatever is in the build\ndirectory, just like Live Server.\n\nCommon use case for Wutch involves writing docs with [Sphinx][sphinx]. Wutch will\nwatch for the changes in all `*.rst` files and automatically rebuild documentation.\nIt will also open a browser window pointing to the build directory and inject\nevery built webpage with a code that will auto-refresh that page after each\nrebuild.\n\n<p>\n    <a href="https://pypi.org/project/wutch/"><img alt="PyPI" src="https://img.shields.io/pypi/v/wutch?color=blue&logo=pypi"></a>\n    <a href=\'https://wutch.readthedocs.io/en/latest/?badge=latest\'><img src=\'https://readthedocs.org/projects/wutch/badge/?version=latest\' alt=\'Documentation Status\' /></a>\n</p>\n\n![Wutch Demo](https://github.com/vduseev/wutch/raw/master/docs/_static/wutch-demo.gif)\n\n## Installation\n\n```shell\npip install wutch\n```\n\n## Usage\n\nJust run wutch in the directory where you want to watch for the changes.\nBy default, `wutch` will:\n\n* Watch for every change in the current directory.\n* Ignore changes in the `_build/` and `build` directories.\n* Run `sphinx-build` shell command for every change in the files.\n* Open a browser pointing to `index.html` in the `_build` directory.\n* Automatically refresh that page every time you change the files\n  and shell command runs.\n\n```shell\n$ wutch\n\n2021-04-02 23:47:06.216 | DEBUG    | wutch.config:__init__:48 - Config{\'dirs\': [\'docs\'], \'ignore_dirs\': [], \'patterns\': [\'*.rst\', \'*.py\'], \'ignore_patterns\': [], \'command\': \'make -C docs rebuild\', \'build\': \'docs/_build/html\', \'inject_patterns\': [\'*.html\'], \'index\': \'index.html\', \'host\': \'localhost\', \'port\': 5010, \'wait\': 3, \'no_browser\': False, \'no_server\': False}\n2021-04-02 23:47:06.217 | DEBUG    | wutch.watcher:start:24 - Starting observer thread\n2021-04-02 23:47:06.219 | DEBUG    | wutch.watcher:start:26 - Observer thred started\n2021-04-02 23:47:06.220 | DEBUG    | wutch.server:start:44 - Server thread started\n2021-04-02 23:47:06.220 | DEBUG    | wutch.server:_open_browser:133 - Opening browser at: http://localhost:5010/index.html\n```\n\nStop wutch by issuing a <kbd>Ctrl+C</kbd> key sequence.\n\n```shell\n^C2021-04-02 23:47:25.283 | DEBUG    | wutch.threaded:run:28 - Stopping all threads on KeyboardInterrupt\n2021-04-02 23:47:25.283 | DEBUG    | wutch.watcher:stop:30 - Stopping observer thread\n2021-04-02 23:47:26.260 | DEBUG    | wutch.watcher:stop:33 - Observer thread stopped\n2021-04-02 23:47:26.260 | DEBUG    | wutch.server:stop:58 - Server thread stopped\n```\n\n## Configuration\n\n### Parameters\n\n```shell\n  -h, --help            show this help message and exit\n  -c COMMAND, --command COMMAND\n                        Shell command executed in response to file changes. Defaults to: sphinx-build.\n  -p [PATTERNS ...], --patterns [PATTERNS ...]\n                        Matches paths with these patterns (separated by \' \'). Defaults to: [\'*\'].\n  -P [IGNORE_PATTERNS ...], --ignore-patterns [IGNORE_PATTERNS ...]\n                        Ignores file changes in these patterns (separated by \' \'). Defaults to: [].\n  -d [DIRS ...], --dirs [DIRS ...]\n                        Directories to watch (separated by \' \'). Defaults to: [\'.\'].\n  -D [IGNORE_DIRS ...], --ignore-dirs [IGNORE_DIRS ...]\n                        Ignore file changes in these directories (separated by \' \'). Defaults to: [\'_build\', \'build\'].\n  -w WAIT, --wait WAIT  Wait N seconds after the command is finished before refreshing the web page. Defaults to: 3.\n  -b BUILD, --build BUILD\n                        Build directory containing files to render in the browser. Defaults to: _build.\n  -I [INJECT_PATTERNS ...], --inject-patterns [INJECT_PATTERNS ...]\n                        Patterns of files to inject with JS code that refreshes them on rebuild (separated by \' \'). Defaults to: [\'*.htm*\'].\n  -i INDEX, --index INDEX\n                        File that will be opened in the browser with the start of the watcher. Defaults to: index.html.\n  --host HOST           Host to bind internal HTTP server to. Defaults to: localhost.\n  --port PORT           TCP port to bind internal HTTP server to. Defaults to: 5010.\n  -B NO_BROWSER, --no-browser NO_BROWSER\n                        Do not open browser at wutch launch. Defaults to: False.\n  -S NO_SERVER, --no-server NO_SERVER\n                        Do not start the webserver, just launch the shell command. Defaults to: False.\n```\n\n### Loading order\n\nWutch loads configuration settings in the following priority:\n\n1. Command line arguments\n2. Environment variables starting with `WUTCH_`\n3. Configuration file `wutch.cfg`\n4. Default variables\n\nEvery variable can be specified in any of the sources above, thanks to\n[`ilexconf`][ilexconf] configuration management library.\n\nFor example, `dirs` variable that lists directories to watch can be\nspecified in several ways:\n\n**Command line:**\n\n```shell\nwutch --dirs . ../other_dir\n```\n\n**Environment variables starting with `WUTCH_`:**\n\n```shell\nexport WUTCH_DIRS=". ../other_dir"\n```\n\n**Configuration file `wutch.cfg`:**\n\n```json\n{\n    "dirs": [".", "../other_dir"]\n}\n```\n\n## Wutch\'s documentation is built using `wutch`\n\nTake a look at the `wutch.cfg` file at the root of the repository. This\nserves as a somewhat common configuration for Sphinx dependent documentation.\n\nWutch documentation is developed using `wutch` and this config below.\n\n```json\n{\n    "dirs": ["docs"],\n    "ignore_dirs": [],\n    "patterns": ["*.rst", "*.py"],\n    "ignore_patterns": [],\n    "command": "make -C docs rebuild",\n    "build": "docs/_build/html",\n    "inject_patterns": ["*.html"],\n    "index": "index.html",\n    "host": "localhost",\n    "port": 5010\n}\n```\n\n\n[sphinx]: https://www.sphinx-doc.org/ "Sphinx"\n[ilexconf]: https://github.com/ilexconf/ilexconf "Ilexconf"',
    'author': 'vduseev',
    'author_email': 'vagiz@duseev.com',
    'maintainer': 'vduseev',
    'maintainer_email': 'vagiz@duseev.com',
    'url': 'https://github.com/vduseev/wutch',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from loguru import logger

from .config import Config
from .server import Server
from .watcher import Watcher
from .events import EventDispatcher
from .threaded import Threaded


class WutchApplication:
    def __init__(self) -> None:

        self.config = None
        self.dispatcher = EventDispatcher()

    def cli(self):

        # Initialize config (will parse arguments and throw SystemExit if
        # something like flag '--help' is used).
        self.config = Config()
        threads = []

        logger.configure()

        watcher = Watcher(self.config, self.dispatcher)
        threads.append(watcher)

        if not self.config.no_server:
            server = Server(self.config, self.dispatcher)
            threads.append(server)

        Threaded.run(threads)
